import ClimaComms
import ArgParse
import CUDA
import ClimaCoupler
import ClimaCoupler: Utilities
import ClimaAtmos as CA
import Random

redirect_stderr(IOContext(stderr, :stacktrace_types_limited => Ref(false)))
Random.seed!(1234)

function argparse_settings()
    s = ArgParse.ArgParseSettings()
    ArgParse.@add_arg_table! s begin
        "--config_file"
        help = "A yaml file used to set the configuration of the coupled model"
        arg_type = String
        default = nothing
    end
    return s
end

function time_to_seconds(s::String)
    factor = Dict("secs" => 1, "mins" => 60, "hours" => 60 * 60, "days" => 60 * 60 * 24)
    s == "Inf" && return Inf
    if count(occursin.(keys(factor), Ref(s))) != 1
        error("Bad format for flag $s. Examples: [`10secs`, `20mins`, `30hours`, `40days`]")
    end
    for match in keys(factor)
        occursin(match, s) || continue
        return parse(Float64, first(split(s, match))) * factor[match]
    end
    error("Uncaught case in computing time from given string.")
end

# Read in atmos configuration file from command line
parse_commandline(s) = ArgParse.parse_args(ARGS, s)
parsed_args = parse_commandline(argparse_settings())
atmos_config_file = parsed_args["config_file"]

# Use input atmos configuration file
if isnothing(atmos_config_file)
    @info "Using Atmos default configuration"
    config = CA.default_config_dict()
else
    @info "Using Atmos configuration from $atmos_config_file"
    config = CA.override_default_config(joinpath(pkgdir(ClimaCoupler), atmos_config_file))
end

# Specify atmos output directory to be inside the coupler output directory
output_dir =
    joinpath(pkgdir(ClimaCoupler), "experiments", "AMIP", "output", "climaatmos", config["job_id"]) * "_artifacts"
!isdir(output_dir) && mkpath(output_dir)
config = merge(config, Dict("output_dir" => output_dir))
atmos_config = CA.AtmosConfig(config)
simulation = CA.get_simulation(atmos_config)
sol_res = CA.solve_atmos!(simulation)

## Use ClimaAtmos calculation to show the simulated years per day of the simulation (SYPD)
tspan = (Float64(0), Float64(time_to_seconds(config["t_end"])))
walltime = sol_res.walltime
es = CA.EfficiencyStats(tspan, walltime)
sypd = CA.simulated_years_per_day(es)
@info "SYPD: $sypd"

## Save the SYPD and allocation information
comms_ctx = atmos_config.comms_ctx
if ClimaComms.iamroot(comms_ctx)
    sypd_filename = joinpath(output_dir, "sypd.txt")
    write(sypd_filename, "$sypd")

    cpu_allocs_GB = Utilities.show_memory_usage(comms_ctx)
    cpu_allocs_filename = joinpath(output_dir, "allocations_cpu.txt")
    write(cpu_allocs_filename, cpu_allocs_GB)
end
